from xflask.type import Enum


class SysCode(Enum):

    SUCCESS                 = '00000', 'Success'
    SYS_ERROR               = '10001', 'Internal Server Error'
    INVALID                 = '10003', 'Request Invalid'
    NOT_FOUND               = '10004', 'Resource Not Found'
    EXISTED                 = '10005', 'Resource Existed'
    AUTH_INVALID            = '10006', 'Authentication Invalid'

    API_NOT_AVAILABLE       = '10007', 'API Not Available'



from flask_jwt_extended import JWTManager, get_jwt_identity, unset_jwt_cookies

from .auth_manager import AuthManager


class JwtAuthManager(AuthManager):

    def init(self, application):
        JWTManager(application.app)

    def get_current_user(self):
        return get_jwt_identity()

    def logout(self):
        unset_jwt_cookies()



from wtforms.validators import *


class EnumValidator(object):

    def __init__(self, enum=None, message=None):
        self.enum = enum
        self.message = message

    def __call__(self, form, field):
        if field.data is not None:
            enum = self.enum.value_of(field.data)
            if enum is None:
                if self.message is None:
                    message = field.gettext('The value is invalid')
                else:
                    message = self.message

                field.errors[:] = []
                raise StopValidation(message)



import logging

from flask import request
from werkzeug.routing import Map, Rule


class Filter(object):

    def init(self, application):
        pass

    def before(self):
        pass

    def after(self, response):
        return response


class ApiLoggingFilter(Filter):

    def __init__(self, routes=None, excludes=None):
        self.routes = routes or ['/api/<path:route>']
        self.excludes = excludes or ['/api/login']

        rules = []
        for route in self.routes:
            rules.append(Rule(route))

        self.matcher = Map(rules).bind('', '/')

        rules = []
        for route in self.excludes:
            rules.append(Rule(route))

        self.excludes_matcher = Map(rules).bind('', '/')

    def init(self, application):
        self._logger = logging.getLogger(self.__class__.__name__)

    def before(self):
        if self.routes is None or len(self.routes) == 0:
            return

        if self.excludes_matcher.test(request.path) is True:
            return

        if self.matcher.test(request.path) is True:
            self._logger.debug('>>> Request')
            self._logger.debug('%s %s' % (request.method, request.url))

            if request.is_json is True:
                try:
                    data = request.get_json()
                except Exception:
                    data = ''

                self._logger.debug(data)

    def after(self, response):
        if self.excludes_matcher.test(request.path) is True:
            return response

        if response.is_json is True:
            self._logger.debug('<<< Response')
            self._logger.debug(response.get_json())

        return response



import os
import xflask


def get_root_dir():
    return os.environ['PROJ_HOME']

def path_exist(path):
    return os.path.exists(path)

def is_dir(path):
    return os.path.isdir(path)

def is_file(path):
    return os.path.isfile(path)

def get_dir_path(path):
    return os.path.dirname(path)

def get_dir_name(path):
    return get_file_name(get_dir_path(path))

def get_file_path(file):
    return os.path.join(get_root_dir(), file)

def get_file_name(path):
    return os.path.basename(path)

def get_xflask_path(path=None):
    xflask_dir = get_dir_path(xflask.__file__)
    return os.path.join(xflask_dir, path)



from flask import request
from flask_jwt_extended import verify_jwt_in_request
from flask_jwt_extended.exceptions import *
from werkzeug.routing import Map, Rule

from xflask.exception import Exception
from xflask.type.sys_code import SysCode
from xflask.web.filter import Filter


class JwtAuthFilter(Filter):

    def __init__(self, open_routes=None):
        self.open_routes = open_routes or ['/api/login', '/static/<path:route>']

        rules = []
        for route in self.open_routes:
            rules.append(Rule(route))

        self.matcher = Map(rules).bind('', '/')

    def before(self):
        if self.open_routes is None or len(self.open_routes) == 0:
            return

        if self.matcher.test(request.path) is False:
            try:
                verify_jwt_in_request()
            except JWTExtendedException:
                raise Exception(SysCode.AUTH_INVALID)



import collections


def get_attr(obj, attr_name):
    return obj[attr_name] if isinstance(obj, dict) else getattr(obj, attr_name)


def to_dict(obj):
    if not hasattr(obj, '__dict__'):
        return obj

    if hasattr(obj, 'to_dict'):
        return obj.to_dict()

    result = {}
    for key, val in obj.__dict__.items():
        if key.startswith("_"):
            continue

        element = []
        if isinstance(val, list):
            for item in val:
                element.append(to_dict(item))
        else:
            element = to_dict(val)

        result[key] = element

    return result


#### DICT ####

def merge_dict(obj_dict: dict, merge_dct: dict):
    for k, v in iter(merge_dct.items()):
        if k in obj_dict and isinstance(obj_dict[k], dict) and isinstance(merge_dct[k], dict):
            merge_dict(obj_dict[k], merge_dct[k])
        else:
            obj_dict[k] = merge_dct[k]


def update_dict(obj_dict, key, value):
    _key = key.split(':')[0]
    _next_key = ''.join(key.split(':')[1:])

    for k, v in obj_dict.items():
        if k == _key:
            if isinstance(v, collections.Mapping):
                update_dict(v, _next_key, value)
            else:
                if type(value) == type(v):
                    obj_dict[_key] = value



import logging

from flask import render_template
from flask import request
from werkzeug.exceptions import NotFound, BadRequest, MethodNotAllowed

from xflask.common.util import to_dict
from xflask.exception import Exception as Sys_Exception
from xflask.type.sys_code import SysCode
from xflask.web.response import Response


class ErrorHandler(object):

    def init(self, application):
        pass

    def handle_404(self, e):
        pass

    def handle_400(self, e):
        pass

    def handle_500(self, e):
        pass


class SimpleErrorHandler(ErrorHandler):
    DEF_API_ROUTE = '/api'
    DEF_TEMPLATE_FOLDER = ''

    def __init__(self, api_route=DEF_API_ROUTE, template_folder=DEF_TEMPLATE_FOLDER):
        self.api_route = api_route
        self.template_folder = template_folder

    def init(self, application):
        self.logger = logging.getLogger(self.__class__.__name__)

        application.app.register_error_handler(NotFound, self.handle_404)
        application.app.register_error_handler(BadRequest, self.handle_400)
        application.app.register_error_handler(MethodNotAllowed, self.handle_400)
        application.app.register_error_handler(Exception, self.handle_500)

    def handle_404(self, e):
        self.logger.exception('404 error')

        if request.path.startswith(self.api_route):
            return to_dict(Response.fail(SysCode.NOT_FOUND))
        else:
            return render_template(self.template_folder + '404.html')

    def handle_400(self, e):
        self.logger.exception('400 error')

        if request.path.startswith(self.api_route):
            return to_dict(Response.fail(SysCode.INVALID))
        else:
            return render_template(self.template_folder + '400.html')

    def handle_500(self, e):
        self.logger.exception('500 error')

        if request.path.startswith(self.api_route):
            if isinstance(e, Sys_Exception):
                response = Response.fail(e.code, e.data)
            else:
                response = Response.fail(SysCode.SYS_ERROR)

            return to_dict(response)
        else:
            data = None
            if isinstance(e, Sys_Exception):
                data = e.data

            return render_template(self.template_folder + '500.html', errors=data)



import os
import subprocess


def run_cmd(command):
    return os.system(command)

def run_cmd_popen(command):
    return os.popen(command)

def run_cmd_subprocess(command, **kwargs):
    return subprocess.Popen(args=command, **kwargs)



import yaml
from yaml import Loader

from xflask.common.util import load_config, get_file_path, merge_dict


class Configuration:

    def __init__(self, conf_file):
        self.cfg = load_config(conf_file)

    def exist(self, key):
        return True if self.get(key) is not None else False

    def get(self, key, default=None):
        value = None
        try:
            keys = key.split(':')
            if len(keys) > 1:
                value = self.__get_nest_value(self.cfg[keys[0]], keys[1:])
            else:
                value = self.cfg[key]

            if value is None and default is not None:
                value = default

        except Exception:
            pass

        return value

    def set(self, key, value):
        self.cfg.__setitem__(key, value)

    def __get_nest_value(self, map_, keys):
        if len(keys) > 1:
            return self.__get_nest_value(map_[keys[0]], keys[1:])
        else:
            return map_[keys[0]]

    def dump(self):
        return yaml.dump(self.cfg, Dumper=yaml.Dumper)

    def merge(self, conf_file):
        with open(get_file_path(conf_file), 'r') as file:
            cfg = yaml.load(file, Loader=Loader)

            merge_dict(self.cfg, cfg)



from datetime import datetime

from flask import json
from sqlalchemy import inspect
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm.attributes import QueryableAttribute

from xflask.sqlalchemy import db
from xflask.web.security import get_current_user


class Model(db.Model):
    __abstract__ = True

    @staticmethod
    def _get_readonly_columns():
        return ['id']

    @classmethod
    def get_column(cls, col_name):
        return cls.__table__.columns.get(col_name)

    @classmethod
    def get_sort_expression(cls, col_name, sort='asc'):
        column = cls.get_column(col_name)
        if column is None:
            return None

        return column.asc() if sort == 'asc' else column.desc()

    def to_dict(self, show=[], _hide=[], _path=None):
        """Return a dictionary representation of this model."""

        hidden = self._hidden_columns if hasattr(self, '_hidden_columns') else []
        hidden = list(set(hidden) | set(_hide) - set(show))

        if not _path:
            _path = self.__tablename__.lower()

            def prepend_path(item):
                item = item.lower()

                if item.split('.', 1)[0] == _path:
                    return item
                if len(item) == 0:
                    return item
                if item[0] != '.':
                    item = '.%s' % item

                item = '%s%s' % (_path, item)

                return item

            show[:] = [prepend_path(x) for x in show]
            _hide[:] = [prepend_path(x) for x in _hide]

        ret_data = {}

        #### columns  ####

        columns = self.__table__.columns.keys()
        for key in columns:
            if key.startswith('_'):
                continue

            check = '%s.%s' % (_path, key)
            if check in _hide or key in hidden:
                continue

            ret_data[key] = getattr(self, key)

        #### relationships ####

        unloaded_relationships = inspect(self).unloaded
        relationships = self.__mapper__.relationships.keys()
        for key in relationships:
            if key.startswith('_'):
                continue

            check = '%s.%s' % (_path, key)
            if check in _hide or key in hidden or key in unloaded_relationships:
                continue

            _hide.append(check)

            is_list = self.__mapper__.relationships[key].uselist
            if is_list:
                items = getattr(self, key)
                if self.__mapper__.relationships[key].query_class is not None:
                    if hasattr(items, 'all'):
                        items = items.all()

                ret_data[key] = []
                for item in items:
                    ret_data[key].append(
                        item.to_dict(
                            show=list(show),
                            _hide=list(_hide),
                            _path=('%s.%s' % (_path, key.lower()))
                        )
                    )
            else:
                if (
                        self.__mapper__.relationships[key].query_class is not None
                        or self.__mapper__.relationships[key].instrument_class is not None
                ):
                    item = getattr(self, key)
                    if item is not None:
                        ret_data[key] = item.to_dict(
                            show=list(show),
                            _hide=list(_hide),
                            _path=('%s.%s' % (_path, key.lower()))
                        )
                    else:
                        ret_data[key] = None
                else:
                    ret_data[key] = getattr(self, key)

        #### properties ####

        properties = dir(self)
        for key in list(set(properties) - set(columns) - set(relationships)):
            if key.startswith('_'):
                continue
            if not hasattr(self.__class__, key):
                continue

            attr = getattr(self.__class__, key)
            if not (isinstance(attr, property) or isinstance(attr, QueryableAttribute)):
                continue

            check = '%s.%s' % (_path, key)
            if check in _hide or key in hidden:
                continue

            val = getattr(self, key)
            if hasattr(val, 'to_dict'):
                ret_data[key] = val.to_dict(
                    show=list(show),
                    _hide=list(_hide),
                    _path=('%s.%s' % (_path, key.lower()))
                )
            else:
                try:
                    ret_data[key] = json.loads(json.dumps(val))
                except Exception:
                    pass

        return ret_data

    def from_dict(self, **kwargs):
        """Update this model with a dictionary."""

        _force = kwargs.pop("_force", False)

        readonly = self._readonly_columns if hasattr(self, '_readonly_columns') else []
        if hasattr(self, '_hidden_columns'):
            readonly += self._hidden_columns

        readonly += self._get_readonly_columns()

        changes = {}

        #### columns ####

        columns = self.__table__.columns.keys()
        for key in columns:
            if key.startswith('_'):
                continue
            allowed = True if _force or key not in readonly else False
            exists = True if key in kwargs else False
            if allowed and exists:
                val = getattr(self, key)
                if val != kwargs[key]:
                    changes[key] = {'old': val, 'new': kwargs[key]}
                    setattr(self, key, kwargs[key])

        #### relationships ####

        relationships = self.__mapper__.relationships.keys()
        for rel in relationships:
            if key.startswith('_'):
                continue
            allowed = True if _force or rel not in readonly else False
            exists = True if rel in kwargs else False
            if allowed and exists:
                is_list = self.__mapper__.relationships[rel].uselist
                if is_list:
                    valid_ids = []
                    query = getattr(self, rel)
                    cls = self.__mapper__.relationships[rel].argument()
                    for item in kwargs[rel]:
                        if (
                                'id' in item
                                and query.filter_by(id=item['id']).limit(1).count() == 1
                        ):
                            obj = cls.query.filter_by(id=item['id']).first()
                            col_changes = obj.from_dict(**item)
                            if col_changes:
                                col_changes['id'] = str(item['id'])
                                if rel in changes:
                                    changes[rel].append(col_changes)
                                else:
                                    changes.update({rel: [col_changes]})
                            valid_ids.append(str(item['id']))
                        else:
                            col = cls()
                            col_changes = col.from_dict(**item)
                            query.append(col)
                            db.session.flush()
                            if col_changes:
                                col_changes['id'] = str(col.id)
                                if rel in changes:
                                    changes[rel].append(col_changes)
                                else:
                                    changes.update({rel: [col_changes]})
                            valid_ids.append(str(col.id))

                    # delete rows from relationship that were not in kwargs[rel]
                    for item in query.filter(not (cls.id.in_(valid_ids))).all():
                        col_changes = {'id': str(item.id), 'deleted': True}
                        if rel in changes:
                            changes[rel].append(col_changes)
                        else:
                            changes.update({rel: [col_changes]})
                        db.session.delete(item)

                else:
                    val = getattr(self, rel)
                    if self.__mapper__.relationships[rel].query_class is not None:
                        if val is not None:
                            col_changes = val.from_dict(**kwargs[rel])
                            if col_changes:
                                changes.update({rel: col_changes})
                    else:
                        if val != kwargs[rel]:
                            setattr(self, rel, kwargs[rel])
                            changes[rel] = {'old': val, 'new': kwargs[rel]}

        #### properties ####

        properties = dir(self)
        for key in list(set(properties) - set(columns) - set(relationships)):
            if key.startswith('_'):
                continue
            allowed = True if _force or key not in readonly else False
            exists = True if key in kwargs else False
            if allowed and exists and getattr(self.__class__, key).fset is not None:
                val = getattr(self, key)
                if hasattr(val, 'to_dict'):
                    val = val.to_dict()
                changes[key] = {'old': val, 'new': kwargs[key]}
                setattr(self, key, kwargs[key])

        return changes


class AuditModel(Model):
    __abstract__ = True

    created_at = db.Column(db.DateTime, default=datetime.now)
    modified_at = db.Column(db.DateTime, onupdate=datetime.now)

    @staticmethod
    def _get_readonly_columns():
        return super()._get_readonly_columns() + ['modified_at', 'created_at', 'modified_by', 'created_by']

    @declared_attr
    def created_by(self):
        return db.Column(db.Integer, default=_current_user_id_or_none)

    @declared_attr
    def modified_by(self):
        return db.Column(db.Integer, onupdate=_current_user_id_or_none)

    # @staticmethod
    # def before_insert(mapper, connection, instance):
    #     user = get_current_user() or {}
    #     instance.created_at = func.now()
    #     instance.created_by = user.get('id')
    #
    # @staticmethod
    # def before_update(mapper, connection, instance):
    #     user = get_current_user() or {}
    #     instance.modified_at = func.now()
    #     instance.modified_by = user.get('id')

    # @classmethod
    # def __declare_last__(cls):
    #     db.event.listen(cls, 'before_insert', cls.before_insert)
    #     db.event.listen(cls, 'before_update', cls.before_update)


class SoftModel(AuditModel):
    __abstract__ = True

    deleted_at = db.Column(db.DateTime)

    @staticmethod
    def _get_readonly_columns(self):
        return super()._get_readonly_columns() + ['deleted_at', 'deleted_by']

    @declared_attr
    def deleted_by(self):
        return db.Column(db.Integer)

    def init_soft_columns(self):
        self.deleted_at = datetime.now()
        self.deleted_by = _current_user_id_or_none()

    # @staticmethod
    # def before_delete(mapper, connection, instance):
    #     user = get_current_user() or {}
    #     instance.deleted_at = func.now()
    #     instance.deleted_by = user.get('id')

    # @classmethod
    # def __declare_last__(cls):
    #     db.event.listen(cls, 'before_delete', cls.before_delete)


def _current_user_id_or_none():
    user = get_current_user() or {}
    return user.get('id')



import sys
import importlib
import re
import yaml

from xflask.common.util.file_util import *
from xflask.common.util.obj_util import *


def setup_env(root_dir_name):
    cur_dir = os.getcwd()
    while True:
        if os.path.basename(cur_dir) == root_dir_name:
            root_dir = cur_dir
            break
        else:
            cur_dir = os.path.dirname(cur_dir)

    os.environ['PROJ_HOME'] = root_dir
    os.environ['XFLASK_HOME'] = os.path.dirname(xflask.__file__)

    sys.path.append(root_dir)

def import_modules(root_dir, model_pkgs):
    for model_pkg in model_pkgs:
        for module in os.listdir(root_dir + '/' + model_pkg.replace('.', '/')):
            if module == '__init__.py' or module[-3:] != '.py':
                continue

            model_namespace = model_pkg + '.' + module[:-3]

            importlib.import_module(model_namespace)

def load_config(conf_file):
    with open(conf_file, 'r') as f:
        param_matcher = re.compile(r'.*\$\{([^}^{]+)\}.*')

        def param_constructor(loader, node):
            value = node.value

            params = param_matcher.findall(value)
            for param in params:
                try:
                    param_value = os.environ[param]
                    return value.replace('${' + param + '}', param_value)
                except Exception:
                    pass

            return value

        class VariableLoader(yaml.SafeLoader):
            pass

        VariableLoader.add_implicit_resolver('!param', param_matcher, None)
        VariableLoader.add_constructor('!param', param_constructor)

        config = yaml.load(f.read(), Loader=VariableLoader)

        return config



class Dto(object):

    def __init__(self, **kwargs):
        for name, value in kwargs.items():
            if isinstance(value, dict):
                value = Dto(**value)

            self.__dict__[name] = value



from xflask.type import Enum
from xflask.type import SysCode


class Exception(Exception):

    def __init__(self, code: Enum = SysCode.SYS_ERROR, data=None):
        self.code = code
        self.data = data



from xflask.wtforms import Field
from xflask.wtforms import ValidationError

class EnumField(Field):

    def __init__(self, enum, label=None, validators=None, **kwargs):
        super(EnumField, self).__init__(label, validators, **kwargs)
        self.enum = enum

    def process_formdata(self, value_list):
        if value_list is None or len(value_list) == 0:
            return None

        self.data = self.enum.value_of(value_list[0])
        if self.data is None:
            raise ValidationError('This field is invalid')

    def _value(self):
        return self.data.code() if self.data is not None else None




import collections
import logging.config
import os

from xflask.common.util import load_config, merge_dict


class Logger(object):

    def __init__(self, conf_file):
        self.config = load_config(conf_file)

        self.init_log_dir(self.config)

        logging.config.dictConfig(self.config)

    def merge(self, conf_file):
        config = load_config(conf_file)
        merge_dict(self.config, config)

        logging.config.dictConfig(self.config)

    def init_log_dir(self, dict_):
        for k, v in dict_.items():
            if k == 'filename':
                filename = v
                file_dir = os.path.dirname(os.path.abspath(filename))

                if not os.path.exists(file_dir):
                    os.makedirs(file_dir)
            elif isinstance(v, collections.Mapping):
                self.init_log_dir(v)



from sqlalchemy import *
from sqlalchemy.orm import *
from flask_sqlalchemy import SQLAlchemy

from .decorator import *
from .util import *


db = SQLAlchemy(session_options={'autocommit': True})

session = db.session


def transactional(subtransactions=True, nested=False):
    def function(f):
        def wrapper(*args, **kwargs):
            db.session.begin(subtransactions=subtransactions, nested=nested)
            try:
                result = f(*args, **kwargs)
                db.session.commit()
            except Exception as e:
                db.session.rollback()
                raise e

            return result

        return wrapper

    return function



from datetime import datetime

from xflask.sqlalchemy.model import Model
from xflask.type.enum import Enum


class JsonSerializer(object):

    def check(self, obj):
        return False

    def serialize(self, obj):
        return None


class EnumSerializer(JsonSerializer):

    def check(self, obj):
        return isinstance(obj, Enum)

    def serialize(self, obj):
        return obj.code()


class DateTimeSerializer(JsonSerializer):

    def __init__(self, fmt='%d-%m-%Y %H:%M:%S'):
        self.fmt = fmt

    def check(self, obj):
        return isinstance(obj, datetime)

    def serialize(self, obj):
        return obj.strftime(self.fmt)


class ModelSerializer(JsonSerializer):

    def check(self, obj):
        return isinstance(obj, Model)

    def serialize(self, obj):
        return obj.to_dict()



class Param(object):
    def __init__(self, name=None, required=True):
        self.name = name
        self.required = required


class Header(object):

    def __init__(self, name=None, required=True):
        self.name = name
        self.required = required



from collections import Mapping, MutableMapping

import six
from wtforms import Form
try:
    from wtforms_sqlalchemy.fields import (
        QuerySelectField,
        QuerySelectMultipleField
    )
    HAS_SQLALCHEMY_SUPPORT = True
except ImportError:
    try:
        from wtforms.ext.sqlalchemy.fields import (
            QuerySelectField,
            QuerySelectMultipleField
        )
        HAS_SQLALCHEMY_SUPPORT = True
    except ImportError:
        HAS_SQLALCHEMY_SUPPORT = False
from wtforms.fields import (
    _unset_value,
    BooleanField,
    Field,
    FieldList,
    FileField,
    FormField,
    StringField
)
from wtforms.validators import DataRequired, Optional

__version__ = '0.3.3'


class InvalidData(Exception):
    pass


def flatten_json(
    form,
    json,
    parent_key='',
    separator='-',
    skip_unknown_keys=True
):
    """Flattens given JSON dict to cope with WTForms dict structure.

    :form form: WTForms Form object
    :param json: json to be converted into flat WTForms style dict
    :param parent_key: this argument is used internally be recursive calls
    :param separator: default separator
    :param skip_unknown_keys:
        if True unknown keys will be skipped, if False throws InvalidData
        exception whenever unknown key is encountered

    Examples::

        >>> flatten_json(MyForm, {'a': {'b': 'c'}})
        {'a-b': 'c'}
    """
    if not isinstance(json, Mapping):
        raise InvalidData(
            u'This function only accepts dict-like data structures.'
        )

    items = []
    for key, value in json.items():
        try:
            unbound_field = getattr(form, key)
        except AttributeError:
            if skip_unknown_keys:
                continue
            else:
                raise InvalidData(u"Unknown field name '%s'." % key)

        try:
            field_class = unbound_field.field_class
        except AttributeError:
            if skip_unknown_keys:
                continue
            else:
                raise InvalidData(u"Key '%s' is not valid field class." % key)

        new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, MutableMapping):
            if issubclass(field_class, FormField):
                nested_form_class = unbound_field.bind(Form(), '').form_class
                items.extend(
                    flatten_json(nested_form_class, value, new_key)
                    .items()
                )
            else:
                items.append((new_key, value))
        elif isinstance(value, list):
            if issubclass(field_class, FieldList):
                nested_unbound_field = unbound_field.bind(
                    Form(),
                    ''
                ).unbound_field
                items.extend(
                    flatten_json_list(
                        nested_unbound_field,
                        value,
                        new_key,
                        separator
                    )
                )
            else:
                items.append((new_key, value))
        else:
            items.append((new_key, value))
    return dict(items)


def flatten_json_list(field, json, parent_key='', separator='-'):
    items = []
    for i, item in enumerate(json):
        new_key = parent_key + separator + str(i)
        if (
            isinstance(item, dict) and
            issubclass(getattr(field, 'field_class'), FormField)
        ):
            nested_class = field.field_class(
                *field.args,
                **field.kwargs
            ).bind(Form(), '').form_class
            items.extend(
                flatten_json(nested_class, item, new_key)
                .items()
            )
        else:
            items.append((new_key, item))
    return items


@property
def patch_data(self):
    if hasattr(self, '_patch_data'):
        return self._patch_data

    data = {}

    def is_optional(field):
        return Optional in [v.__class__ for v in field.validators]

    def is_required(field):
        return DataRequired in [v.__class__ for v in field.validators]

    for name, f in six.iteritems(self._fields):
        if f.is_missing:
            if is_optional(f):
                continue
            elif not is_required(f) and f.default is None:
                continue
            elif isinstance(f, FieldList) and f.min_entries == 0:
                continue

        if isinstance(f, FormField):
            data[name] = f.patch_data
        elif isinstance(f, FieldList):
            if issubclass(f.unbound_field.field_class, FormField):
                data[name] = [i.patch_data for i in f.entries]
            else:
                data[name] = [i.data for i in f.entries]
        else:
            data[name] = f.data
    return data


def monkey_patch_field_process(func):
    """
    Monkey patches Field.process method to better understand missing values.
    """
    def process(self, formdata, data=_unset_value):
        call_original_func = True
        if not isinstance(self, FormField):

            if formdata and self.name in formdata:
                if (
                    len(formdata.getlist(self.name)) == 1 and
                    formdata.getlist(self.name) == [None]
                ):
                    call_original_func = False
                    self.data = None
                self.is_missing = not bool(formdata.getlist(self.name))
            else:
                self.is_missing = True

        if call_original_func:
            func(self, formdata, data=data)

        if (
            formdata and self.name in formdata and
            formdata.getlist(self.name) == [None] and
            isinstance(self, FormField)
        ):
            self.form._is_missing = False
            self.form._patch_data = None

        if isinstance(self, StringField) and not isinstance(self, FileField):
            if not self.data:
                try:
                    self.data = self.default()
                except TypeError:
                    self.data = self.default
            else:
                self.data = six.text_type(self.data)

    return process


class MultiDict(dict):
    def getlist(self, key):
        val = self[key]
        if not isinstance(val, list):
            val = [val]
        return val

    def getall(self, key):
        return [self[key]]


@classmethod
def from_json(
    cls,
    formdata=None,
    obj=None,
    prefix='',
    data=None,
    meta=None,
    skip_unknown_keys=True,
    **kwargs
):
    form = cls(
        formdata=MultiDict(
            flatten_json(cls, formdata, skip_unknown_keys=skip_unknown_keys)
        ) if formdata else None,
        obj=obj,
        prefix=prefix,
        data=data,
        meta=meta,
        **kwargs
    )
    return form


@property
def is_missing(self):
    if hasattr(self, '_is_missing'):
        return self._is_missing

    for name, field in self._fields.items():
        if not field.is_missing:
            return False
    return True


@property
def field_list_is_missing(self):
    if hasattr(self, '_is_missing'):
        return self._is_missing

    return all([field.is_missing for field in self.entries])


def monkey_patch_process_formdata(func):
    def process_formdata(self, valuelist):
        valuelist = list(map(six.text_type, valuelist))

        return func(self, valuelist)
    return process_formdata


def init():
    Form.is_missing = is_missing
    FieldList.is_missing = field_list_is_missing
    Form.from_json = from_json
    Form.patch_data = patch_data
    FieldList.patch_data = patch_data
    if HAS_SQLALCHEMY_SUPPORT:
        QuerySelectField.process_formdata = monkey_patch_process_formdata(
            QuerySelectField.process_formdata
        )
        QuerySelectMultipleField.process_formdata = \
            monkey_patch_process_formdata(
                QuerySelectMultipleField.process_formdata
            )
    Field.process = monkey_patch_field_process(Field.process)
    FormField.process = monkey_patch_field_process(FormField.process)
    BooleanField.false_values = BooleanField.false_values + (False,)



import inspect
import logging
import sys

from flask import Flask
from flask.json import JSONEncoder
from flask_injector import FlaskInjector, request
from flask_sqlalchemy import SQLAlchemy
from injector import singleton as singleton
from werkzeug.utils import find_modules, import_string

from xflask.common.configuration import Configuration
from xflask.common.logger import Logger
from xflask.common.util import get_root_dir, get_file_path, get_xflask_path, import_modules
from xflask.component import Component
from xflask.json import EnumSerializer, DateTimeSerializer, ModelSerializer
from xflask.web.controller import Controller
from xflask.web.error_handler import SimpleErrorHandler
from xflask.web.filter import Filter
from xflask.web.security.jwt_auth_filter import JwtAuthFilter
from xflask.web.security.jwt_auth_manager import JwtAuthManager


class Application(object):
    DEF_CONF_FILE = 'conf/server.yml'
    DEF_LOG_FILE = 'conf/logging.yml'

    DEF_FILTERS = [JwtAuthFilter()]
    DEF_ERROR_HANDLER = SimpleErrorHandler()
    DEF_AUTH_MANAGER = JwtAuthManager()

    DEF_JSON_SERIALIZERS = [EnumSerializer(), DateTimeSerializer(), ModelSerializer()]

    def __init__(self, db, conf_file=None, filters=None, error_handler=None, auth_manager=None, json_serializers=[],
                 listeners=[]):

        self.db = db
        self.conf_file = conf_file or get_file_path('main/conf/server.yml')

        self.filters = filters or self.DEF_AUTH_MANAGER
        self.error_handler = error_handler or self.DEF_ERROR_HANDLER
        self.auth_manager = auth_manager or self.DEF_AUTH_MANAGER
        self.json_serializers = json_serializers + self.DEF_JSON_SERIALIZERS
        self.listeners = listeners
        self._pre_init()

    def run(self, host=None, port=None):
        host = host or self.conf.get('HOST')
        port = port or self.conf.get('PORT')
        self.app.run(host, port, use_reloader=False)

    def _pre_init(self):
        self._init_config()

        self._init_logging()

        self._init_app()

        self._init_db()

        # register xflask extension
        if not hasattr(self.app, 'extensions'):
            self.app.extensions = {}

        self.app.extensions['xflask'] = self

    def init(self):
        # before init listener
        for listener in self.listeners:
            listener.before_init(self)

        self._init_error_handler()

        self._register_models()

        self._register_components()

        self._register_controllers()

        self._register_blueprints()

        self._register_filters()

        self._init_auth_manager()

        # after init listener
        for listener in self.listeners:
            listener.after_init(self)

    def _init_config(self):
        self.conf = Configuration(get_xflask_path(self.DEF_CONF_FILE))
        if self.conf_file is not None:
            self.conf.merge(self.conf_file)

    def _init_logging(self):
        if self.conf.get('LOGGING') is False:
            logging.basicConfig(level=logging.DEBUG)
        else:
            self.logger = Logger(get_xflask_path(self.DEF_LOG_FILE))

            log_file = self.conf.get('LOGGING_CONF_FILE')
            if log_file is not None:
                self.logger.merge(log_file)

        self.logger = logging.getLogger(self.__class__.__name__)

    def _init_app(self):
        self.app = Flask(self.conf.get('APP_NAME'),
                         static_folder=self.conf.get('STATIC_DIR'),
                         template_folder=self.conf.get('TEMPLATE_DIR'),
                         root_path=get_root_dir())

        # config
        self.app.config.from_mapping(self.conf.cfg)

        # json encoder
        json_serializers = self.json_serializers

        class _JsonEncoder(JSONEncoder):

            def default(self, obj):
                for serializer in json_serializers:
                    if serializer.check(obj):
                        return serializer.serialize(obj)

                return super().default(obj)

        self.app.json_encoder = _JsonEncoder

    def _init_db(self):
        if self.conf.exist('SQLALCHEMY_DATABASE_URI'):
            self.db.init_app(self.app)

    def _init_auth_manager(self):
        if self.auth_manager is None:
            return

        self.auth_manager.init(self)

    def _init_error_handler(self):
        self.error_handler.init(self)

    def _register_filters(self):
        for filter_ in self.filters:
            if isinstance(filter_, Filter):
                filter_.init(self)
                self.app.before_request(filter_.before)
                self.app.after_request(filter_.after)
            else:
                self.logger.debug('!!! Invalid filter: %s', filter_.__name__)

    def _register_blueprints(self):
        for package in self.conf.get('BLUEPRINT_PKG'):
            try:
                for name in find_modules(package):
                    try:
                        module = import_string(name)
                        if hasattr(module, 'Blueprint'):
                            self.app.register_blueprint(module.bp)
                        else:
                            del module

                    except Exception as e:
                        if not isinstance(e, AttributeError):
                            self.logger.error('!!! Failed to register blueprint: ', e)
            except Exception as e:
                self.logger.error('!!! Failed to find blueprint module in package: %s', package, e)

    def _register_models(self):
        import_modules(get_root_dir(), self.conf.get('MODEL_PKG') or [])

    def get_component(self, clazz):
        return self.flask_injector.injector.get(clazz)

    def _register_components(self):

        def configure(binder):
            # config
            binder.bind(Configuration, self.conf, scope=singleton)

            # db
            if self.conf.exist('SQLALCHEMY_DATABASE_URI'):
                binder.bind(SQLAlchemy, to=self.db, scope=singleton)

            # component
            for package in self.conf.get('COMPONENT_PKG'):
                try:
                    self.logger.debug('>>> Scan modules in %s', package)

                    for pkg_name in find_modules(package):
                        try:
                            module = import_string(pkg_name)
                            class_names = [m[0] for m in inspect.getmembers(module, inspect.isclass) if
                                           m[1].__module__ == module.__name__]

                            valid_module = False
                            for class_name in class_names:
                                obj = getattr(module, class_name)

                                if issubclass(obj, Component):
                                    if obj.abstract is True:
                                        continue

                                    valid_module = True

                                    scope = singleton if obj.scope == 'singleton' else request
                                    binder.bind(obj, obj, scope)

                                    self.logger.debug('Register component: %s', class_name)
                                else:
                                    del module
                                    self.logger.debug('!!! Invalid component: %s', class_name)

                            if not valid_module:
                                del module

                        except Exception as e:
                            self.logger.exception('!!! Failed to initialize component in %s', pkg_name, e)
                            sys.exit()

                except Exception:
                    self.logger.debug('!!! No modules founded in %s', package)

        self.flask_injector = FlaskInjector(app=self.app, modules=[configure])

    def _register_controllers(self):
        for package in self.conf.get('CONTROLLER_PKG'):
            try:
                self.logger.debug('>>> Scan modules in %s', package)

                for module_ns in find_modules(package):
                    try:
                        module = import_string(module_ns)
                        class_names = [m[0] for m in inspect.getmembers(module, inspect.isclass) if
                                       m[1].__module__ == module.__name__]

                        valid_module = False
                        for class_name in class_names:
                            controller = getattr(module, class_name)

                            if issubclass(controller, Controller):
                                if controller.abstract is True:
                                    continue

                                valid_module = True

                                try:
                                    controller.register(self.app, self.flask_injector.injector)
                                except Exception:
                                    self.logger.exception('Failed to initialize controller: %s', class_name)
                                    sys.exit()

                                self.logger.debug('Register controller: %s', class_name)
                            else:
                                self.logger.debug('!!! Invalid controller: %s', class_name)

                        if not valid_module:
                            del module

                    except Exception as e:
                        self.logger.exception('!!! Failed to initialize controller in %s', module_ns, e)
                        sys.exit()

            except Exception:
                self.logger.debug('!!! No modules found in %s', package)

        # display registered routes
        self.logger.debug('>>> Registered routes:')

        routes = self.app.url_map._rules
        if routes is not None and len(routes) > 0:
            max_len = max([len(route.rule) for route in routes])
            for route in routes:
                self.logger.debug('%*s | %26s | %s', max_len, route.rule, route.methods, route.endpoint)



from xflask.component import Component
from xflask.sqlalchemy import session, transactional


class Dao(Component):

    def __init__(self, model):
        self.model = model

    def count(self, **criterion):
        return self.query().filter_by(**criterion).count()

    def exist(self, id=None, **criterion):
        query = self.query().filter_by(**self._get_pk_criterion(id)) \
            if id is not None else self.query().filter_by(**criterion)
        return query.scalar() is not None

    def get(self, id=None, **criterion):
        return self.query().filter_by(**self._get_pk_criterion(id)).first() \
            if id is not None else self.query().filter_by(**criterion).first()

    def get_all(self, order=(), **criterion):
        return self.query().filter_by(**criterion).order_by(*order).all()

    def get_page(self, page=1, per_page=30, order=(), **criterion):
        criterion = filter_criterion(**criterion)

        return self.query().filter_by(**criterion).order_by(*order).paginate(page, per_page=per_page)

    def query(self, *models):
        if models is None or len(models) == 0:
            return session.query(self.model)
        else:
            return session.query(*models)

    @transactional()
    def insert(self, obj):
        if isinstance(obj, dict):
            obj = self.model(**obj)

        session.add(obj)

    @transactional()
    def update(self, obj, **criterion):
        if len(criterion) > 0:
            self.query().filter_by(**criterion).update(obj if isinstance(obj, dict) else obj.to_dict())
        elif isinstance(obj, dict):
            self.query().filter_by(**self._get_pk_criterion_by_object(obj)).update(obj)
        else:
            self._merge(obj)

    @transactional()
    def delete(self, obj=None, **criterion):
        if len(criterion) > 0:
            self.query().filter_by(**criterion).delete()
        elif obj is not None:
            if isinstance(obj, (int, float)) or isinstance(obj, tuple):
                self.query().filter_by(**self._get_pk_criterion_by_object(obj)).delete()
            else:
                session.delete(obj)

    def delete_all(self):
        self.query().delete()

    def _begin(self, subtransactions=True, nested=False):
        session.begin(subtransactions=subtransactions, nested=nested)

    def _begin_nested(self):
        session.begin_nested()

    def _flush(self, objs):
        session.flush(objs)

    def _merge(self, obj):
        return session.merge(obj)

    def _commit(self):
        session.commit()

    def _rollback(self):
        session.rollback()

    def _get_pk_criterion(self, id):
        if isinstance(id, (int, float)):
            id = (id,)

        criterion = {}
        for idx, pk in enumerate(self.model.__mapper__.primary_key):
            criterion[pk.name] = id[idx]

        return criterion

    def _get_pk_criterion_by_object(self, obj):
        criterion = {}
        for idx, pk in enumerate(self.model.__mapper__.primary_key):
            criterion[pk.name] = obj[pk.name]

        return criterion

def filter_criterion(**criterion):
    ret_dict = {}
    for field, value in criterion.items():
        if value is not None:
            ret_dict[field] = value
    return ret_dict



import sys
from flask import current_app


def get_xflask():
    return get_extension('xflask')

def get_extension(extension):
    try:
        return current_app.extensions.get(extension)
    except KeyError:
        raise RuntimeError("You must initialize xflask before using this method")

def exit(code=None):
    def function(f):
        def wrapper(*args, **kwargs):
            try:
                result = f(*args, **kwargs)
            except Exception:
                sys.exit(code)
            return result
        return wrapper
    return function



from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()

def requirement():
    return [
        'Flask',
        'Flask-Script',
        'Flask-Migrate',
        'Flask-SQLAlchemy',
        'Flask-Testing',
        'Flask-Injector',
        'Flask-JWT-Extended',
        'Flask-WTF',
        'WTForms',
        'email-validator',
        'Werkzeug==0.16.1',
        'PyYAML',
        'psycopg2-binary'
    ]

setup(
    name='xFlask',
    packages=find_packages(),
    package_data={'xflask': ['conf/*']},
    version='0.1.7',
    description='Python Web Framework',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='LEANG Sotheara',
    author_email='leangsotheara@gmail.com',
    url='https://github.com/sotheara-leang/xFlask.git',
    keywords=['xFlask', 'Restful', 'Web'],
    install_requires=requirement(),
    python_requires=">=3.5",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License'
    ]
)



class Page(object):

    def __init__(self, page=None, per_page=None, total_page=None, total=None, items=None):
        self.page = page
        self.per_page = per_page
        self.total_page = total_page
        self.total = total
        self.items = items

    @staticmethod
    def from_pagination(obj):
        return Page(page=obj.page, per_page=obj.per_page, total_page=obj.pages, total=obj.total, items=obj.items)



from flask_sqlalchemy.model import DefaultMeta

from xflask.component import Component
from xflask.dao import Dao


class Service(Component):
    ...


class CrudService(Service):

    def __init__(self, dao: Dao):
        if isinstance(dao, DefaultMeta):
            self.dao = Dao(dao)
        elif isinstance(dao, Dao):
            self.dao = dao
        else:
            raise Exception('Invalid dao')

    def count(self, **criterion):
        return self.dao.count(**criterion)

    def exist(self, id=None, **criterion):
        return self.dao.exist(id, **criterion)

    def get(self, id=None, **criterion):
        return self.dao.get(id, **criterion)

    def get_all(self, **criterion):
        return self.dao.get_all(**criterion)

    def get_page(self, page, sort, criterion):
        return self.dao.get_page(page.page, page.per_page, sort, **criterion)

    def query(self, *models):
        return self.dao.query(*models)

    def create(self, obj):
        self.dao.insert(obj)

    def update(self, obj, **criterion):
        self.dao.update(obj, **criterion)

    def delete(self, obj=None, **criterion):
        self.dao.delete(obj, **criterion)

    def delete_all(self):
        self.dao.delete_all()

    def _begin(self, subtransactions=True, nested=False):
        self.dao.begin(subtransactions=subtransactions, nested=nested)

    def _begin_nested(self):
        self.dao.begin_nested()

    def _flush(self, objs):
        self.dao.flush(objs)

    def _merge(self, obj):
        return self.dao.merge(obj)

    def _commit(self):
        self.dao.commit()

    def _rollback(self):
        self.dao.rollback()



from wtforms import Form

from xflask.dto import Dto
from xflask.wtforms import StringField, IntegerField, FieldList, FormField, SelectField


class Form(Form):
    class Meta:
        csrf = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        exclude = kwargs.get('exclude') or []
        self.exclude = exclude

    def get_dto(self):
        return Dto(**self.data)


class SortForm(Form):
    field = StringField()
    order = SelectField(choices=[('asc', ''), ('desc', '')])


class PageForm(Form):
    page = IntegerField(default=1)
    per_page = IntegerField(default=30)
    sort = FieldList(FormField(SortForm), default=[])

    def get_page(self):
        return Dto(page=self.page.data, per_page=self.per_page.data)

    def get_sort(self):
        ret_data = []
        for sort in self.sort:
            ret_data.append(Dto(field=sort.field.data, order=sort.order.data))
        return ret_data

    def get_criterion(self):
        criterion = {}
        for field in self:
            field_name = field.id or field.name

            if field_name not in ['page', 'per_page', 'sort']:
                criterion[field_name] = field.data

        return Dto(**criterion)
